/**
 * 
 * Ulfius Framework changed simple example program
 * 
 * This example program described the main features 
 * that are available in a callback function.
 * 
 * All changes are for complete controlling of steps1(2,3) programs
 * 
 * Copyright 2019 Luděk Horáček <12@12in.cz>
 * Copyright 2015-2017 Nicolas Mora <mail@babelouest.org>
 * 
 * License GPL
 *
 */

#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/signal.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include "../../src/ulfius.h"

#define SFORWARD "/forward"
#define SBACKWARD "/backward"
#define SPOS "/pos"
#define SZERO "/zero"
#define SSTOP "/stop"
#define SSHUTDOWN "/shutdown"

#ifndef SELAXIS
#define SELAXIS "axis1.h"
#endif

#include SELAXIS

/**
 * callback functions declaration
 */
int callback_get_forward (const struct _u_request * request, struct _u_response * response, void * user_data);
int callback_get_backward (const struct _u_request * request, struct _u_response * response, void * user_data);
int callback_get_pos (const struct _u_request * request, struct _u_response * response, void * user_data);
int callback_get_zero (const struct _u_request * request, struct _u_response * response, void * user_data);
int callback_get_stop (const struct _u_request * request, struct _u_response * response, void * user_data);
int callback_get_shutdown (const struct _u_request * request, struct _u_response * response, void * user_data);
int callback_default (const struct _u_request * request, struct _u_response * response, void * user_data);

pid_t  pid;
FILE *fptr;
char *ptr;

char * read_file(const char * filename) {
  char * buffer = NULL;
  long length;
  FILE * f = fopen (filename, "rb");
  if (filename != NULL) {

    if (f) {
      fseek (f, 0, SEEK_END);
      length = ftell (f);
      fseek (f, 0, SEEK_SET);
      buffer = o_malloc (length + 1);
      if (buffer) {
        fread (buffer, 1, length, f);
      }
      buffer[length] = '\0';
      fclose (f);
    }
    return buffer;
  } else {
    return NULL;
  }
}

int main (int argc, char **argv) {
  int ret;
  unlink(STEPSID);
  
  // Set the framework port number
  struct _u_instance instance;
  
  y_init_logs("HTTPREST", Y_LOG_MODE_CONSOLE, Y_LOG_LEVEL_DEBUG, NULL, "Starting HTTPREST");
  
  if (ulfius_init_instance(&instance, PORT, NULL, NULL) != U_OK) {
    y_log_message(Y_LOG_LEVEL_ERROR, "Error ulfius_init_instance, abort");
    return(1);
  }
  
  u_map_put(instance.default_headers, "Access-Control-Allow-Origin", "*");
  
  // Maximum body size sent by the client is 1 Kb
  instance.max_post_body_size = 1024;
  
  // Endpoint list declaration
  ulfius_add_endpoint_by_val(&instance, "GET", SFORWARD, "/:steps/:speed/:acceleration", 0, &callback_get_forward, NULL);
  ulfius_add_endpoint_by_val(&instance, "GET", SBACKWARD, "/:steps/:speed/:acceleration", 0, &callback_get_backward, NULL);
  
  ulfius_add_endpoint_by_val(&instance, "GET", SPOS, NULL, 0, &callback_get_pos, NULL);
  ulfius_add_endpoint_by_val(&instance, "GET", SZERO, NULL, 0, &callback_get_zero, NULL);
  ulfius_add_endpoint_by_val(&instance, "GET", SSTOP, NULL, 0, &callback_get_stop, NULL);
  ulfius_add_endpoint_by_val(&instance, "GET", SSHUTDOWN, NULL, 0, &callback_get_shutdown, NULL);
  
  // default_endpoint declaration
  ulfius_set_default_endpoint(&instance, &callback_default, NULL);
  
  // Start the framework
  if (argc == 4 && strcmp("-secure", argv[1]) == 0) {
    // If command-line options are -secure <key_file> <cert_file>, then open an https connection
    char * key_pem = read_file(argv[2]), * cert_pem = read_file(argv[3]);
    ret = ulfius_start_secure_framework(&instance, key_pem, cert_pem);
    o_free(key_pem);
    o_free(cert_pem);
  } else {
    // Open an http connection
    ret = ulfius_start_framework(&instance);
  }
  
  if (ret == U_OK) {
    y_log_message(Y_LOG_LEVEL_DEBUG, "Start %sframework on port %d", ((argc == 4 && strcmp("-secure", argv[1]) == 0)?"secure ":""), instance.port);
    
    // Wait for the user to press <E> on the console to quit the application
    while (69 != getchar());
  } else {
    y_log_message(Y_LOG_LEVEL_DEBUG, "Error starting framework");
  }
  y_log_message(Y_LOG_LEVEL_DEBUG, "End framework");
  
  y_close_logs();
  
  ulfius_stop_framework(&instance);
  ulfius_clean_instance(&instance);
  
  return 0;
}

/**
 * Forward function
 */
int callback_get_forward (const struct _u_request * request, struct _u_response * response, void * user_data) {
 const char * steps = u_map_get(request->map_url, "steps");
 const char * speed = u_map_get(request->map_url, "speed");
 const char * acceleration = u_map_get(request->map_url, "acceleration");
 
 if (access(STEPSID, F_OK ) != -1) { // only one stepping process is possible
  ulfius_set_string_body_response(response, 200, "NO");
  return U_CALLBACK_CONTINUE;
 }
 else {
  pid = fork();
  if (pid == 0) {
   execlp(STEPSPROG, SUDOSTEPSPROG, steps, "0", speed, acceleration, NULL);
  }
 }

 ulfius_set_string_body_response(response, 200, "OK");
 return U_CALLBACK_CONTINUE;
}

/**
 * Backward function
 */
int callback_get_backward (const struct _u_request * request, struct _u_response * response, void * user_data) {
 const char * steps = u_map_get(request->map_url, "steps");
 const char * speed = u_map_get(request->map_url, "speed");
 const char * acceleration = u_map_get(request->map_url, "acceleration");

 if (access(STEPSID, F_OK ) != -1) { // only one stepping process is possible
  ulfius_set_string_body_response(response, 200, "NO");
  return U_CALLBACK_CONTINUE;
 }
 else {
  pid = fork();
  if (pid == 0) {
   execlp(STEPSPROG, SUDOSTEPSPROG, steps, "1", speed, acceleration, NULL);
  }
 }

 ulfius_set_string_body_response(response, 200, "OK");
 return U_CALLBACK_CONTINUE;
}

/**
 * STOP function
 */
int callback_get_stop (const struct _u_request * request, struct _u_response * response, void * user_data) {

 if (access(STEPSID, F_OK ) != -1) { // send SIGTERM to existing stepping process
  char cfch[256];
  fptr = fopen(STEPSID, "r");
  fgets(cfch, 256, fptr);
  fclose(fptr);

  pid = strtol(cfch, &ptr, 10);
  kill(pid, SIGTERM);
 }

 ulfius_set_string_body_response(response, 200, "OK");
 return U_CALLBACK_CONTINUE;
}

/**
 * Get position function
 */
int callback_get_pos (const struct _u_request * request, struct _u_response * response, void * user_data) {
 char cfc[256], cfch[256];
 unlink(POSACT);

 if (access(STEPSID, F_OK ) != -1) { // relative position get from fresh temporary file
  fptr = fopen(STEPSID, "r");
  fgets(cfc, 256, fptr);
  fclose(fptr);

  pid = strtol(cfc, &ptr, 10);
  kill(pid, SIGUSR1);

  usleep(250);
  while (access(POSACT, R_OK) != 0) usleep(50);

  fptr = fopen(POSACT, "r");
  fgets(cfch, 256, fptr);
  fclose(fptr);
 }
 else { // get relative position from regular file
  fptr = fopen(POSTXT, "r");
  fgets(cfch, 256, fptr);
  fclose(fptr);
 }

 ulfius_set_string_body_response(response, 200, cfch); // return position in response body
 return U_CALLBACK_CONTINUE;
}

/**
 * Set zero position function
 */
int callback_get_zero (const struct _u_request * request, struct _u_response * response, void * user_data) {
 char cfch[1] = "0";

 // reset relative position in regular file
 fptr = fopen(POSTXT, "w");
 fwrite(&cfch, sizeof(cfch), 1, fptr); 
 fclose(fptr);

 ulfius_set_string_body_response(response, 200, "OK");
 return U_CALLBACK_CONTINUE;
}

/**
 * Shutdown function
 */
int callback_get_shutdown (const struct _u_request * request, struct _u_response * response, void * user_data) {

 if (access(STEPSID, F_OK ) != -1) { // send SIGTERM to existing stepping process
  char cfch[256];
  fptr = fopen(STEPSID, "r");
  fgets(cfch, 256, fptr);
  fclose(fptr);

  pid = strtol(cfch, &ptr, 10);
  kill(pid, SIGTERM);
 }

 system("sudo shutdown"); // shutdown Rapberry Pi

 ulfius_set_string_body_response(response, 200, "OK");
 return U_CALLBACK_CONTINUE;
}

/**
 * Default callback function called if no endpoint has a match
 */
int callback_default (const struct _u_request * request, struct _u_response * response, void * user_data) {
  ulfius_set_string_body_response(response, 404, "Page not found");
  return U_CALLBACK_CONTINUE;
}
